! Code here has been copied from fv_mapz.F90; this is just a lightweight
! module used for vertical interpolation in coarse_graining.F90.
module mappm_mod

  implicit none
  private

  real, parameter:: r3 = 1./3., r23 = 2./3., r12 = 1./12.

  public :: mappm

  contains

  !>@brief The subroutine 'mappm' is a general-purpose routine for remapping
  !! one set of vertical levels to another.  This is copied from fv_mapz.F90
  !! to prevent the need for coarse_graining.F90 to depend on the dynamical
  !! core.
  subroutine mappm(km, pe1, q1, kn, pe2, q2, i1, i2, iv, kord, ptop)

    ! IV = 0: constituents
    ! IV = 1: potential temp
    ! IV =-1: winds
     
    ! Mass flux preserving mapping: q1(im,km) -> q2(im,kn)
     
    ! pe1: pressure at layer edges (from model top to bottom surface)
    !      in the original vertical coordinate
    ! pe2: pressure at layer edges (from model top to bottom surface)
    !      in the new vertical coordinate
    
     integer, intent(in):: i1, i2, km, kn, kord, iv
     real, intent(in ):: pe1(i1:i2,km+1), pe2(i1:i2,kn+1) !< pe1: pressure at layer edges from model top to bottom
                                                          !!      surface in the ORIGINAL vertical coordinate 
                                                          !< pe2: pressure at layer edges from model top to bottom 
                                                          !!      surface in the NEW vertical coordinate
    ! Mass flux preserving mapping: q1(im,km) -> q2(im,kn)
     real, intent(in )::  q1(i1:i2,km)
     real, intent(out)::  q2(i1:i2,kn)
     real, intent(IN) :: ptop
    ! local
          real  qs(i1:i2)
          real dp1(i1:i2,km)
          real a4(4,i1:i2,km)
          integer i, k, l
          integer k0, k1
          real pl, pr, tt, delp, qsum, dpsum, esl
    
          do k=1,km
             do i=i1,i2
                 dp1(i,k) = pe1(i,k+1) - pe1(i,k)
                a4(1,i,k) = q1(i,k)
             enddo
          enddo
    
          if ( kord >7 ) then
               call  cs_profile( qs, a4, dp1, km, i1, i2, iv, kord )
          else
               call ppm_profile( a4, dp1, km, i1, i2, iv, kord )
          endif
    
    !------------------------------------
    ! Lowest layer: constant distribution
    !------------------------------------
#ifdef NGGPS_SUBMITTED
          do i=i1,i2
             a4(2,i,km) = q1(i,km)
             a4(3,i,km) = q1(i,km)
             a4(4,i,km) = 0.
          enddo
#endif
    
          do 5555 i=i1,i2
             k0 = 1
          do 555 k=1,kn
    
             if(pe2(i,k) .le. pe1(i,1)) then
    ! above old ptop
                q2(i,k) = q1(i,1)
             elseif(pe2(i,k) .ge. pe1(i,km+1)) then
    ! Entire grid below old ps
#ifdef NGGPS_SUBMITTED
                q2(i,k) = a4(3,i,km)   ! this is not good.
#else
                q2(i,k) = q1(i,km)
#endif
             else
    
             do 45 L=k0,km
    ! locate the top edge at pe2(i,k)
             if( pe2(i,k) .ge. pe1(i,L) .and.        &
                 pe2(i,k) .le. pe1(i,L+1)    ) then
                 k0 = L
                 PL = (pe2(i,k)-pe1(i,L)) / dp1(i,L)
                 if(pe2(i,k+1) .le. pe1(i,L+1)) then
    
    ! entire new grid is within the original grid
                   PR = (pe2(i,k+1)-pe1(i,L)) / dp1(i,L)
                   TT = r3*(PR*(PR+PL)+PL**2)
                   q2(i,k) = a4(2,i,L) + 0.5*(a4(4,i,L)+a4(3,i,L)  &
                           - a4(2,i,L))*(PR+PL) - a4(4,i,L)*TT
                  goto 555
                 else
    ! Fractional area...
                  delp = pe1(i,L+1) - pe2(i,k)
                  TT   = r3*(1.+PL*(1.+PL))
                  qsum = delp*(a4(2,i,L)+0.5*(a4(4,i,L)+            &
                         a4(3,i,L)-a4(2,i,L))*(1.+PL)-a4(4,i,L)*TT)
                  dpsum = delp
                  k1 = L + 1
                 goto 111
                 endif
             endif
    45       continue
    
    111      continue
             do 55 L=k1,km
             if( pe2(i,k+1) .gt. pe1(i,L+1) ) then
    
    ! Whole layer..
    
                qsum  =  qsum + dp1(i,L)*q1(i,L)
                dpsum = dpsum + dp1(i,L)
             else
               delp = pe2(i,k+1)-pe1(i,L)
               esl  = delp / dp1(i,L)
               qsum = qsum + delp * (a4(2,i,L)+0.5*esl*            &
                     (a4(3,i,L)-a4(2,i,L)+a4(4,i,L)*(1.-r23*esl)) )
              dpsum = dpsum + delp
               k0 = L
               goto 123
             endif
    55       continue
            delp = pe2(i,k+1) - pe1(i,km+1)
            if(delp > 0.) then
    ! Extended below old ps
#ifdef NGGPS_SUBMITTED
               qsum = qsum + delp * a4(3,i,km)    ! not good.
#else
               qsum = qsum + delp * q1(i,km)
#endif
              dpsum = dpsum + delp
            endif
    123     q2(i,k) = qsum / dpsum
          endif
    555   continue
    5555  continue
    
    end subroutine mappm
 
    subroutine cs_profile(qs, a4, delp, km, i1, i2, iv, kord)
     ! Optimized vertical profile reconstruction:
     ! Latest: Apr 2008 S.-J. Lin, NOAA/GFDL
      integer, intent(in):: i1, i2
      integer, intent(in):: km      !< vertical dimension
      integer, intent(in):: iv      !< iv =-1: winds
                                    !< iv = 0: positive definite scalars
                                    !< iv = 1: others
      integer, intent(in):: kord
      real, intent(in)   ::   qs(i1:i2)
      real, intent(in)   :: delp(i1:i2,km)     !< layer pressure thickness
      real, intent(inout):: a4(4,i1:i2,km)     !< Interpolated values
     !-----------------------------------------------------------------------
      logical, dimension(i1:i2,km):: extm, ext5, ext6
      real  gam(i1:i2,km)
      real    q(i1:i2,km+1)
      real   d4(i1:i2)
      real   bet, a_bot, grat 
      real   pmp_1, lac_1, pmp_2, lac_2, x0, x1
      integer i, k, im
     
      if ( iv .eq. -2 ) then
           do i=i1,i2
              gam(i,2) = 0.5
                q(i,1) = 1.5*a4(1,i,1)
           enddo
           do k=2,km-1
              do i=i1, i2
                       grat = delp(i,k-1) / delp(i,k)
                        bet =  2. + grat + grat - gam(i,k)
                     q(i,k) = (3.*(a4(1,i,k-1)+a4(1,i,k)) - q(i,k-1))/bet
                 gam(i,k+1) = grat / bet
              enddo
           enddo
           do i=i1,i2
                 grat = delp(i,km-1) / delp(i,km) 
              q(i,km) = (3.*(a4(1,i,km-1)+a4(1,i,km)) - grat*qs(i) - q(i,km-1)) /  &
                        (2. + grat + grat - gam(i,km))
              q(i,km+1) = qs(i)
           enddo
           do k=km-1,1,-1
             do i=i1,i2
                q(i,k) = q(i,k) - gam(i,k+1)*q(i,k+1)
             enddo
           enddo
      else
       do i=i1,i2
              grat = delp(i,2) / delp(i,1)   ! grid ratio
               bet = grat*(grat+0.5)
            q(i,1) = ( (grat+grat)*(grat+1.)*a4(1,i,1) + a4(1,i,2) ) / bet
          gam(i,1) = ( 1. + grat*(grat+1.5) ) / bet
       enddo
     
       do k=2,km
          do i=i1,i2
                d4(i) = delp(i,k-1) / delp(i,k)
                  bet =  2. + d4(i) + d4(i) - gam(i,k-1)
               q(i,k) = ( 3.*(a4(1,i,k-1)+d4(i)*a4(1,i,k)) - q(i,k-1) )/bet
             gam(i,k) = d4(i) / bet
          enddo
       enddo
      
       do i=i1,i2
              a_bot = 1. + d4(i)*(d4(i)+1.5)
          q(i,km+1) = (2.*d4(i)*(d4(i)+1.)*a4(1,i,km)+a4(1,i,km-1)-a_bot*q(i,km))  &
                    / ( d4(i)*(d4(i)+0.5) - a_bot*gam(i,km) )
       enddo
     
       do k=km,1,-1
          do i=i1,i2
             q(i,k) = q(i,k) - gam(i,k)*q(i,k+1)
          enddo
       enddo
      endif
     !----- Perfectly linear scheme --------------------------------
      if ( abs(kord) > 16 ) then
       do k=1,km
          do i=i1,i2
             a4(2,i,k) = q(i,k  )
             a4(3,i,k) = q(i,k+1)
             a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
          enddo
       enddo
       return
      endif
     !----- Perfectly linear scheme --------------------------------
     
     !------------------
     ! Apply constraints
     !------------------
       im = i2 - i1 + 1
     
     ! Apply *large-scale* constraints 
       do i=i1,i2
          q(i,2) = min( q(i,2), max(a4(1,i,1), a4(1,i,2)) )
          q(i,2) = max( q(i,2), min(a4(1,i,1), a4(1,i,2)) )
       enddo
     
       do k=2,km
          do i=i1,i2
             gam(i,k) = a4(1,i,k) - a4(1,i,k-1)
          enddo
       enddo
     
     ! Interior:
       do k=3,km-1
          do i=i1,i2
             if ( gam(i,k-1)*gam(i,k+1)>0. ) then
     ! Apply large-scale constraint to ALL fields if not local max/min
                  q(i,k) = min( q(i,k), max(a4(1,i,k-1),a4(1,i,k)) )
                  q(i,k) = max( q(i,k), min(a4(1,i,k-1),a4(1,i,k)) )
             else
               if ( gam(i,k-1) > 0. ) then
     ! There exists a local max
                    q(i,k) = max(q(i,k), min(a4(1,i,k-1),a4(1,i,k)))
               else
     ! There exists a local min
                      q(i,k) = min(q(i,k), max(a4(1,i,k-1),a4(1,i,k)))
                    if ( iv==0 ) q(i,k) = max(0., q(i,k))
               endif
             endif
          enddo
       enddo
     
     ! Bottom:
       do i=i1,i2
          q(i,km) = min( q(i,km), max(a4(1,i,km-1), a4(1,i,km)) )
          q(i,km) = max( q(i,km), min(a4(1,i,km-1), a4(1,i,km)) )
       enddo
     
       do k=1,km
          do i=i1,i2
             a4(2,i,k) = q(i,k  )
             a4(3,i,k) = q(i,k+1)
          enddo
       enddo
     
       do k=1,km
          if ( k==1 .or. k==km ) then
            do i=i1,i2
               extm(i,k) = (a4(2,i,k)-a4(1,i,k)) * (a4(3,i,k)-a4(1,i,k)) > 0.
            enddo
          else
            do i=i1,i2
               extm(i,k) = gam(i,k)*gam(i,k+1) < 0.
            enddo
          endif
          if ( abs(kord) > 9 ) then
            do i=i1,i2
               x0 = 2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k))
               x1 = abs(a4(2,i,k)-a4(3,i,k))
               a4(4,i,k) = 3.*x0
               ext5(i,k) = abs(x0) > x1
               ext6(i,k) = abs(a4(4,i,k)) > x1
            enddo
          endif
       enddo
     
     !---------------------------
     ! Apply subgrid constraints:
     !---------------------------
     ! f(s) = AL + s*[(AR-AL) + A6*(1-s)]         ( 0 <= s  <= 1 )
     ! Top 2 and bottom 2 layers always use monotonic mapping
     
       if ( iv==0 ) then
          do i=i1,i2
             a4(2,i,1) = max(0., a4(2,i,1))
          enddo
       elseif ( iv==-1 ) then 
           do i=i1,i2
              if ( a4(2,i,1)*a4(1,i,1) <= 0. ) a4(2,i,1) = 0.
           enddo
       elseif ( iv==2 ) then
          do i=i1,i2
             a4(2,i,1) = a4(1,i,1)
             a4(3,i,1) = a4(1,i,1)
             a4(4,i,1) = 0.
          enddo
       endif
     
       if ( iv/=2 ) then
          do i=i1,i2
             a4(4,i,1) = 3.*(2.*a4(1,i,1) - (a4(2,i,1)+a4(3,i,1)))
          enddo
          call cs_limiters(im, extm(i1,1), a4(1,i1,1), 1)
       endif
     
     ! k=2
        do i=i1,i2
           a4(4,i,2) = 3.*(2.*a4(1,i,2) - (a4(2,i,2)+a4(3,i,2)))
        enddo
        call cs_limiters(im, extm(i1,2), a4(1,i1,2), 2)
     
     !-------------------------------------
     ! Huynh's 2nd constraint for interior:
     !-------------------------------------
       do k=3,km-2
          if ( abs(kord)<9 ) then
            do i=i1,i2
     ! Left  edges
               pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
               lac_1 = pmp_1 + 1.5*gam(i,k+2)
               a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),   &
                                              max(a4(1,i,k), pmp_1, lac_1) )
     ! Right edges
               pmp_2 = a4(1,i,k) + 2.*gam(i,k)
               lac_2 = pmp_2 - 1.5*gam(i,k-1)
               a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),    &
                                              max(a4(1,i,k), pmp_2, lac_2) )
     
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
     
          elseif ( abs(kord)==9 ) then
            do i=i1,i2
               if ( extm(i,k) .and. extm(i,k-1) ) then  ! c90_mp122
     ! grid-scale 2-delta-z wave detected
                    a4(2,i,k) = a4(1,i,k)
                    a4(3,i,k) = a4(1,i,k)
                    a4(4,i,k) = 0.
               else if ( extm(i,k) .and. extm(i,k+1) ) then  ! c90_mp122
     ! grid-scale 2-delta-z wave detected
                    a4(2,i,k) = a4(1,i,k)
                    a4(3,i,k) = a4(1,i,k)
                    a4(4,i,k) = 0.
               else
                 a4(4,i,k) = 6.*a4(1,i,k) - 3.*(a4(2,i,k)+a4(3,i,k))
     ! Check within the smooth region if subgrid profile is non-monotonic
                 if( abs(a4(4,i,k)) > abs(a4(2,i,k)-a4(3,i,k)) ) then
                       pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                       lac_1 = pmp_1 + 1.5*gam(i,k+2)
                   a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),  &
                                                  max(a4(1,i,k), pmp_1, lac_1) )
                       pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                       lac_2 = pmp_2 - 1.5*gam(i,k-1)
                   a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),  &
                                                  max(a4(1,i,k), pmp_2, lac_2) )
                   a4(4,i,k) = 6.*a4(1,i,k) - 3.*(a4(2,i,k)+a4(3,i,k))
                 endif
               endif
            enddo
          elseif ( abs(kord)==10 ) then
            do i=i1,i2
               if( ext5(i,k) ) then
                   if( ext5(i,k-1) .or. ext5(i,k+1) ) then
                        a4(2,i,k) = a4(1,i,k)
                        a4(3,i,k) = a4(1,i,k)
                   elseif ( ext6(i,k-1) .or. ext6(i,k+1) ) then
                        pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                        lac_1 = pmp_1 + 1.5*gam(i,k+2)
                        a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),  &
                                                       max(a4(1,i,k), pmp_1, lac_1) )
                        pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                        lac_2 = pmp_2 - 1.5*gam(i,k-1)
                        a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),  &
                                                       max(a4(1,i,k), pmp_2, lac_2) )
                   endif
               elseif( ext6(i,k) ) then
                   if( ext5(i,k-1) .or. ext5(i,k+1) ) then
                       pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                       lac_1 = pmp_1 + 1.5*gam(i,k+2)
                       a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),  &
                                                      max(a4(1,i,k), pmp_1, lac_1) )
                       pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                       lac_2 = pmp_2 - 1.5*gam(i,k-1)
                       a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),  &
                                                      max(a4(1,i,k), pmp_2, lac_2) )
                   endif
               endif
            enddo
            do i=i1,i2
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
          elseif ( abs(kord)==12 ) then
            do i=i1,i2
               if( extm(i,k) ) then
     ! grid-scale 2-delta-z wave detected
                   a4(2,i,k) = a4(1,i,k)
                   a4(3,i,k) = a4(1,i,k)
                   a4(4,i,k) = 0.
               else        ! not a local extremum
                 a4(4,i,k) = 6.*a4(1,i,k) - 3.*(a4(2,i,k)+a4(3,i,k))
     ! Check within the smooth region if subgrid profile is non-monotonic
                 if( abs(a4(4,i,k)) > abs(a4(2,i,k)-a4(3,i,k)) ) then
                       pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                       lac_1 = pmp_1 + 1.5*gam(i,k+2)
                   a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),  &
                                                  max(a4(1,i,k), pmp_1, lac_1) )
                       pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                       lac_2 = pmp_2 - 1.5*gam(i,k-1)
                   a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),  &
                                                  max(a4(1,i,k), pmp_2, lac_2) )
                   a4(4,i,k) = 6.*a4(1,i,k) - 3.*(a4(2,i,k)+a4(3,i,k))
                 endif
               endif
            enddo
          elseif ( abs(kord)==13 ) then
            do i=i1,i2
               if( ext6(i,k) ) then
                  if ( ext6(i,k-1) .and. ext6(i,k+1) ) then
     ! grid-scale 2-delta-z wave detected
                      a4(2,i,k) = a4(1,i,k)
                      a4(3,i,k) = a4(1,i,k)
                  endif
               endif
            enddo
            do i=i1,i2
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
          elseif ( abs(kord)==14 ) then
     
            do i=i1,i2
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
     
          elseif ( abs(kord)==15 ) then   ! revised kord=9 scehem
            do i=i1,i2
               if ( ext5(i,k) ) then  ! c90_mp122
                  if ( ext5(i,k-1) .or. ext5(i,k+1) ) then  ! c90_mp122
     ! grid-scale 2-delta-z wave detected
                       a4(2,i,k) = a4(1,i,k)
                       a4(3,i,k) = a4(1,i,k)
                  endif
               elseif( ext6(i,k) ) then
     ! Check within the smooth region if subgrid profile is non-monotonic
                       pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                       lac_1 = pmp_1 + 1.5*gam(i,k+2)
                   a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),  &
                                                  max(a4(1,i,k), pmp_1, lac_1) )
                       pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                       lac_2 = pmp_2 - 1.5*gam(i,k-1)
                   a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),  &
                                                  max(a4(1,i,k), pmp_2, lac_2) )
               endif
            enddo
            do i=i1,i2
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
          elseif ( abs(kord)==16 ) then
            do i=i1,i2
               if( ext5(i,k) ) then
                  if ( ext5(i,k-1) .or. ext5(i,k+1) ) then
                      a4(2,i,k) = a4(1,i,k)
                      a4(3,i,k) = a4(1,i,k)
                  elseif ( ext6(i,k-1) .or. ext6(i,k+1) ) then
                      ! Left  edges
                      pmp_1 = a4(1,i,k) - 2.*gam(i,k+1)
                      lac_1 = pmp_1 + 1.5*gam(i,k+2)
                      a4(2,i,k) = min(max(a4(2,i,k), min(a4(1,i,k), pmp_1, lac_1)),   &
                                          max(a4(1,i,k), pmp_1, lac_1) )
                      ! Right edges
                      pmp_2 = a4(1,i,k) + 2.*gam(i,k)
                      lac_2 = pmp_2 - 1.5*gam(i,k-1)
                      a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), pmp_2, lac_2)),    &
                                          max(a4(1,i,k), pmp_2, lac_2) )
                  endif
               endif
            enddo
            do i=i1,i2
               a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
            enddo
          else      ! kord = 11
            do i=i1,i2
              if ( ext5(i,k) .and. (ext5(i,k-1) .or. ext5(i,k+1)) ) then
     ! Noisy region:
                   a4(2,i,k) = a4(1,i,k)
                   a4(3,i,k) = a4(1,i,k)
                   a4(4,i,k) = 0.
              else
                   a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
              endif
            enddo
          endif
     
     ! Additional constraint to ensure positivity
          if ( iv==0 ) call cs_limiters(im, extm(i1,k), a4(1,i1,k), 0)
     
       enddo      ! k-loop
     
     !----------------------------------
     ! Bottom layer subgrid constraints:
     !----------------------------------
       if ( iv==0 ) then
          do i=i1,i2
             a4(3,i,km) = max(0., a4(3,i,km))
          enddo
       elseif ( iv .eq. -1 ) then 
           do i=i1,i2
              if ( a4(3,i,km)*a4(1,i,km) <= 0. )  a4(3,i,km) = 0.
           enddo
       endif
     
       do k=km-1,km
          do i=i1,i2
             a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
          enddo
          if(k==(km-1)) call cs_limiters(im, extm(i1,k), a4(1,i1,k), 2)
          if(k== km   ) call cs_limiters(im, extm(i1,k), a4(1,i1,k), 1)
       enddo
     
    end subroutine cs_profile
 
    subroutine cs_limiters(im, extm, a4, iv)
      integer, intent(in) :: im
      integer, intent(in) :: iv
      logical, intent(in) :: extm(im)
      real , intent(inout) :: a4(4,im)   !< PPM array
     ! LOCAL VARIABLES:
      real  da1, da2, a6da
      integer i
     
      if ( iv==0 ) then
     ! Positive definite constraint
         do i=1,im
         if( a4(1,i)<=0.) then
             a4(2,i) = a4(1,i)
             a4(3,i) = a4(1,i)
             a4(4,i) = 0.
         else
           if( abs(a4(3,i)-a4(2,i)) < -a4(4,i) ) then
              if( (a4(1,i)+0.25*(a4(3,i)-a4(2,i))**2/a4(4,i)+a4(4,i)*r12) < 0. ) then
     ! local minimum is negative
                  if( a4(1,i)<a4(3,i) .and. a4(1,i)<a4(2,i) ) then
                      a4(3,i) = a4(1,i)
                      a4(2,i) = a4(1,i)
                      a4(4,i) = 0.
                  elseif( a4(3,i) > a4(2,i) ) then
                      a4(4,i) = 3.*(a4(2,i)-a4(1,i))
                      a4(3,i) = a4(2,i) - a4(4,i)
                  else
                      a4(4,i) = 3.*(a4(3,i)-a4(1,i))
                      a4(2,i) = a4(3,i) - a4(4,i)
                  endif
              endif
           endif
         endif
         enddo
      elseif ( iv==1 ) then
         do i=1,im
           if( (a4(1,i)-a4(2,i))*(a4(1,i)-a4(3,i))>=0. ) then
              a4(2,i) = a4(1,i)
              a4(3,i) = a4(1,i)
              a4(4,i) = 0.
           else
              da1  = a4(3,i) - a4(2,i)
              da2  = da1**2
              a6da = a4(4,i)*da1
              if(a6da < -da2) then
                 a4(4,i) = 3.*(a4(2,i)-a4(1,i))
                 a4(3,i) = a4(2,i) - a4(4,i)
              elseif(a6da > da2) then
                 a4(4,i) = 3.*(a4(3,i)-a4(1,i))
                 a4(2,i) = a4(3,i) - a4(4,i)
              endif
           endif
         enddo
      else
     ! Standard PPM constraint
         do i=1,im
           if( extm(i) ) then
              a4(2,i) = a4(1,i)
              a4(3,i) = a4(1,i)
              a4(4,i) = 0.
           else
              da1  = a4(3,i) - a4(2,i)
              da2  = da1**2
              a6da = a4(4,i)*da1
              if(a6da < -da2) then
                 a4(4,i) = 3.*(a4(2,i)-a4(1,i))
                 a4(3,i) = a4(2,i) - a4(4,i)
              elseif(a6da > da2) then
                 a4(4,i) = 3.*(a4(3,i)-a4(1,i))
                 a4(2,i) = a4(3,i) - a4(4,i)
              endif
           endif
         enddo
      endif
    end subroutine cs_limiters
     
    subroutine ppm_profile(a4, delp, km, i1, i2, iv, kord)
     
     ! INPUT PARAMETERS:
      integer, intent(in):: iv      !< iv =-1: winds iv = 0: positive definite scalars iv = 1: others iv = 2: temp (if remap_t) and w (iv=-2)
      integer, intent(in):: i1      !< Starting longitude
      integer, intent(in):: i2      !< Finishing longitude
      integer, intent(in):: km      !< Vertical dimension
      integer, intent(in):: kord    !< Order (or more accurately method no.):
                                    ! 
      real , intent(in):: delp(i1:i2,km)     !< Layer pressure thickness
     
     ! !INPUT/OUTPUT PARAMETERS:
      real , intent(inout):: a4(4,i1:i2,km)  !< Interpolated values
     
     ! DESCRIPTION:
     !
     !   Perform the piecewise parabolic reconstruction
     ! 
     ! !REVISION HISTORY: 
     ! S.-J. Lin   revised at GFDL 2007
     !-----------------------------------------------------------------------
     ! local arrays:
           real    dc(i1:i2,km)
           real    h2(i1:i2,km)
           real  delq(i1:i2,km)
           real   df2(i1:i2,km)
           real    d4(i1:i2,km)
     
     ! local scalars:
           integer i, k, km1, lmt, it
           real  fac
           real  a1, a2, c1, c2, c3, d1, d2
           real  qm, dq, lac, qmp, pmp
     
           km1 = km - 1
            it = i2 - i1 + 1
     
           do k=2,km
              do i=i1,i2
                 delq(i,k-1) =   a4(1,i,k) - a4(1,i,k-1)
                   d4(i,k  ) = delp(i,k-1) + delp(i,k)
              enddo
           enddo
     
           do k=2,km1
              do i=i1,i2
                      c1  = (delp(i,k-1)+0.5*delp(i,k))/d4(i,k+1)
                      c2  = (delp(i,k+1)+0.5*delp(i,k))/d4(i,k)
                 df2(i,k) = delp(i,k)*(c1*delq(i,k) + c2*delq(i,k-1)) /      &
                                         (d4(i,k)+delp(i,k+1))
                 dc(i,k) = sign( min(abs(df2(i,k)),              &
                                 max(a4(1,i,k-1),a4(1,i,k),a4(1,i,k+1))-a4(1,i,k),  &
                       a4(1,i,k)-min(a4(1,i,k-1),a4(1,i,k),a4(1,i,k+1))), df2(i,k) )
              enddo
           enddo
     
     !-----------------------------------------------------------
     ! 4th order interpolation of the provisional cell edge value
     !-----------------------------------------------------------
     
           do k=3,km1
              do i=i1,i2
                 c1 = delq(i,k-1)*delp(i,k-1) / d4(i,k)
                 a1 = d4(i,k-1) / (d4(i,k) + delp(i,k-1))
                 a2 = d4(i,k+1) / (d4(i,k) + delp(i,k))
                 a4(2,i,k) = a4(1,i,k-1) + c1 + 2./(d4(i,k-1)+d4(i,k+1)) *    &
                           ( delp(i,k)*(c1*(a1 - a2)+a2*dc(i,k-1)) -          &
                             delp(i,k-1)*a1*dc(i,k  ) )
              enddo
           enddo
     
     !     if(km>8 .and. kord>4) call steepz(i1, i2, km, a4, df2, dc, delq, delp, d4)
     
     ! Area preserving cubic with 2nd deriv. = 0 at the boundaries
     ! Top
           do i=i1,i2
              d1 = delp(i,1)
              d2 = delp(i,2)
              qm = (d2*a4(1,i,1)+d1*a4(1,i,2)) / (d1+d2)
              dq = 2.*(a4(1,i,2)-a4(1,i,1)) / (d1+d2)
              c1 = 4.*(a4(2,i,3)-qm-d2*dq) / ( d2*(2.*d2*d2+d1*(d2+3.*d1)) )
              c3 = dq - 0.5*c1*(d2*(5.*d1+d2)-3.*d1*d1)
              a4(2,i,2) = qm - 0.25*c1*d1*d2*(d2+3.*d1)
     ! Top edge:
     !-------------------------------------------------------
              a4(2,i,1) = d1*(2.*c1*d1**2-c3) + a4(2,i,2)
     !-------------------------------------------------------
     !        a4(2,i,1) = (12./7.)*a4(1,i,1)-(13./14.)*a4(1,i,2)+(3./14.)*a4(1,i,3)
     !-------------------------------------------------------
     ! No over- and undershoot condition
              a4(2,i,2) = max( a4(2,i,2), min(a4(1,i,1), a4(1,i,2)) )
              a4(2,i,2) = min( a4(2,i,2), max(a4(1,i,1), a4(1,i,2)) )
              dc(i,1) =  0.5*(a4(2,i,2) - a4(1,i,1))
           enddo
     
     ! Enforce monotonicity  within the top layer
     
           if( iv==0 ) then
              do i=i1,i2
                 a4(2,i,1) = max(0., a4(2,i,1))
                 a4(2,i,2) = max(0., a4(2,i,2))
              enddo 
           elseif( iv==-1 ) then
              do i=i1,i2
                 if ( a4(2,i,1)*a4(1,i,1) <= 0. ) a4(2,i,1) = 0.
              enddo
           elseif( abs(iv)==2 ) then
              do i=i1,i2
                 a4(2,i,1) = a4(1,i,1)
                 a4(3,i,1) = a4(1,i,1)
              enddo
           endif
     
     ! Bottom
     ! Area preserving cubic with 2nd deriv. = 0 at the surface
           do i=i1,i2
              d1 = delp(i,km)
              d2 = delp(i,km1)
              qm = (d2*a4(1,i,km)+d1*a4(1,i,km1)) / (d1+d2)
              dq = 2.*(a4(1,i,km1)-a4(1,i,km)) / (d1+d2)
              c1 = (a4(2,i,km1)-qm-d2*dq) / (d2*(2.*d2*d2+d1*(d2+3.*d1)))
              c3 = dq - 2.0*c1*(d2*(5.*d1+d2)-3.*d1*d1)
              a4(2,i,km) = qm - c1*d1*d2*(d2+3.*d1)
     ! Bottom edge:
     !-----------------------------------------------------
              a4(3,i,km) = d1*(8.*c1*d1**2-c3) + a4(2,i,km)
     !        dc(i,km) = 0.5*(a4(3,i,km) - a4(1,i,km))
     !-----------------------------------------------------
     !        a4(3,i,km) = (12./7.)*a4(1,i,km)-(13./14.)*a4(1,i,km-1)+(3./14.)*a4(1,i,km-2)
     ! No over- and under-shoot condition
              a4(2,i,km) = max( a4(2,i,km), min(a4(1,i,km), a4(1,i,km1)) )
              a4(2,i,km) = min( a4(2,i,km), max(a4(1,i,km), a4(1,i,km1)) )
              dc(i,km) = 0.5*(a4(1,i,km) - a4(2,i,km))
           enddo
     
     
     ! Enforce constraint on the "slope" at the surface
#ifdef BOT_MONO
           do i=i1,i2
              a4(4,i,km) = 0
              if( a4(3,i,km) * a4(1,i,km) <= 0. ) a4(3,i,km) = 0.
              d1 = a4(1,i,km) - a4(2,i,km)
              d2 = a4(3,i,km) - a4(1,i,km)
              if ( d1*d2 < 0. ) then
                   a4(2,i,km) = a4(1,i,km)
                   a4(3,i,km) = a4(1,i,km)
              else
                   dq = sign(min(abs(d1),abs(d2),0.5*abs(delq(i,km-1))), d1)
                   a4(2,i,km) = a4(1,i,km) - dq
                   a4(3,i,km) = a4(1,i,km) + dq
              endif
           enddo
#else
           if( iv==0 ) then
               do i=i1,i2
                  a4(2,i,km) = max(0.,a4(2,i,km))
                  a4(3,i,km) = max(0.,a4(3,i,km))
               enddo
           elseif( iv<0 ) then
               do i=i1,i2
                  if( a4(1,i,km)*a4(3,i,km) <= 0. )  a4(3,i,km) = 0.
               enddo
           endif
#endif
     
        do k=1,km1
           do i=i1,i2
              a4(3,i,k) = a4(2,i,k+1)
           enddo
        enddo
     
     !-----------------------------------------------------------
     ! f(s) = AL + s*[(AR-AL) + A6*(1-s)]         ( 0 <= s  <= 1 )
     !-----------------------------------------------------------
     ! Top 2 and bottom 2 layers always use monotonic mapping
           do k=1,2
              do i=i1,i2
                 a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
              enddo
              call ppm_limiters(dc(i1,k), a4(1,i1,k), it, 0)
           enddo
     
           if(kord >= 7) then
     !-----------------------
     ! Huynh's 2nd constraint
     !-----------------------
           do k=2,km1
              do i=i1,i2
     ! Method#1
     !           h2(i,k) = delq(i,k) - delq(i,k-1)
     ! Method#2 - better
                 h2(i,k) = 2.*(dc(i,k+1)/delp(i,k+1) - dc(i,k-1)/delp(i,k-1))  &
                          / ( delp(i,k)+0.5*(delp(i,k-1)+delp(i,k+1)) )        &
                          * delp(i,k)**2 
     ! Method#3
     !!!            h2(i,k) = dc(i,k+1) - dc(i,k-1)
              enddo
           enddo
     
           fac = 1.5           ! original quasi-monotone
     
           do k=3,km-2
             do i=i1,i2
     ! Right edges
     !        qmp   = a4(1,i,k) + 2.0*delq(i,k-1)
     !        lac   = a4(1,i,k) + fac*h2(i,k-1) + 0.5*delq(i,k-1)
     !
              pmp   = 2.*dc(i,k)
              qmp   = a4(1,i,k) + pmp
              lac   = a4(1,i,k) + fac*h2(i,k-1) + dc(i,k)
              a4(3,i,k) = min(max(a4(3,i,k), min(a4(1,i,k), qmp, lac)),    &
                                             max(a4(1,i,k), qmp, lac) )
     ! Left  edges
     !        qmp   = a4(1,i,k) - 2.0*delq(i,k)
     !        lac   = a4(1,i,k) + fac*h2(i,k+1) - 0.5*delq(i,k)
     !
              qmp   = a4(1,i,k) - pmp
              lac   = a4(1,i,k) + fac*h2(i,k+1) - dc(i,k)
              a4(2,i,k) = min(max(a4(2,i,k),  min(a4(1,i,k), qmp, lac)),   &
                          max(a4(1,i,k), qmp, lac))
     !-------------
     ! Recompute A6
     !-------------
              a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
             enddo
     ! Additional constraint to ensure positivity when kord=7
              if (iv == 0 .and. kord >= 6 )                      &
                  call ppm_limiters(dc(i1,k), a4(1,i1,k), it, 2)
           enddo
     
           else
     
              lmt = kord - 3
              lmt = max(0, lmt)
              if (iv == 0) lmt = min(2, lmt)
     
              do k=3,km-2
                 if( kord /= 4) then
                   do i=i1,i2
                      a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
                   enddo
                 endif
                 if(kord/=6) call ppm_limiters(dc(i1,k), a4(1,i1,k), it, lmt)
              enddo
           endif
     
           do k=km1,km
              do i=i1,i2
                 a4(4,i,k) = 3.*(2.*a4(1,i,k) - (a4(2,i,k)+a4(3,i,k)))
              enddo
              call ppm_limiters(dc(i1,k), a4(1,i1,k), it, 0)
           enddo
     
    end subroutine ppm_profile
     
    subroutine ppm_limiters(dm, a4, itot, lmt)
     
     ! INPUT PARAMETERS:
           real , intent(in):: dm(*)     !< Linear slope
           integer, intent(in) :: itot      !< Total Longitudes
           integer, intent(in) :: lmt       !< 0: Standard PPM constraint 1: Improved full monotonicity constraint
                                            !< (Lin) 2: Positive definite constraint 
                                            !< 3: do nothing (return immediately)
     ! INPUT/OUTPUT PARAMETERS:
           real , intent(inout) :: a4(4,*)   !< PPM array AA <-- a4(1,i) AL <-- a4(2,i) AR <-- a4(3,i) A6 <-- a4(4,i)
     ! LOCAL VARIABLES:
           real  qmp
           real  da1, da2, a6da
           real  fmin
           integer i
     
     ! Developer: S.-J. Lin
     
           if ( lmt == 3 ) return
     
           if(lmt == 0) then
     ! Standard PPM constraint
           do i=1,itot
           if(dm(i) == 0.) then
              a4(2,i) = a4(1,i)
              a4(3,i) = a4(1,i)
              a4(4,i) = 0.
           else
              da1  = a4(3,i) - a4(2,i)
              da2  = da1**2
              a6da = a4(4,i)*da1
              if(a6da < -da2) then
                 a4(4,i) = 3.*(a4(2,i)-a4(1,i))
                 a4(3,i) = a4(2,i) - a4(4,i)
              elseif(a6da > da2) then
                 a4(4,i) = 3.*(a4(3,i)-a4(1,i))
                 a4(2,i) = a4(3,i) - a4(4,i)
              endif
           endif
           enddo
     
           elseif (lmt == 1) then
     
     ! Improved full monotonicity constraint (Lin 2004)
     ! Note: no need to provide first guess of A6 <-- a4(4,i)
           do i=1, itot
                qmp = 2.*dm(i)
              a4(2,i) = a4(1,i)-sign(min(abs(qmp),abs(a4(2,i)-a4(1,i))), qmp)
              a4(3,i) = a4(1,i)+sign(min(abs(qmp),abs(a4(3,i)-a4(1,i))), qmp)
              a4(4,i) = 3.*( 2.*a4(1,i) - (a4(2,i)+a4(3,i)) )
           enddo
     
           elseif (lmt == 2) then
     
     ! Positive definite constraint
           do i=1,itot
           if( abs(a4(3,i)-a4(2,i)) < -a4(4,i) ) then
           fmin = a4(1,i)+0.25*(a4(3,i)-a4(2,i))**2/a4(4,i)+a4(4,i)*r12
              if( fmin < 0. ) then
              if(a4(1,i)<a4(3,i) .and. a4(1,i)<a4(2,i)) then
                 a4(3,i) = a4(1,i)
                 a4(2,i) = a4(1,i)
                 a4(4,i) = 0.
              elseif(a4(3,i) > a4(2,i)) then
                 a4(4,i) = 3.*(a4(2,i)-a4(1,i))
                 a4(3,i) = a4(2,i) - a4(4,i)
              else
                 a4(4,i) = 3.*(a4(3,i)-a4(1,i))
                 a4(2,i) = a4(3,i) - a4(4,i)
              endif
              endif
           endif
           enddo
     
           endif
     
    end subroutine ppm_limiters

end module mappm_mod
